# косяк, мой подсчет метрик не работает если там нет трушных 1
"""
CDSDSDS
"""

from sklearn.model_selection import train_test_split
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import torch
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def ts_train_test_split(df, len_seq, 
                     points_ahead=1, gap=0, shag=1, intersection=True,
                     test_size=None,train_size=None, random_state=None, shuffle=True,stratify=None):
    """
    Функция которая разбивает временной ряд на трейн и тест выборки 
    
    Временной ряд здесь это вообще вся история
    Функционал позволяет разбивать ....

    Parameters
    ----------
    df : pd.DataFrame
        Array of shape (n_samples, n_features) with timestamp index.
    len_seq : int 
        Длина 
    flag : bool (optional, default=True)
        If true, then do one thing.
        If false, then do another thing.
    f : callable (optional, default=None)
        Call-back function.  If not specified, then some other function
        will be used
    **kwargs :
        Additional keyword arguments will be passed to name_of_function
    
    Returns
    -------
    z : ndarray
        result of shape (n_samples,).  Note that here we use "ndarray" rather
        than "array_like", because we assure we'll return a numpy array.
    
    Examples
    --------
    >>> X = np.ones((4, 3))
    >>> y = np.ones(4)
    >>> sklearn_template(X, y)
    (z, xmin, xmax)  # this should match the actual output
    """

#    """
#	df - требование, но если тебе не хочется то просто сделай np.array на выходе и все
#    Разбить временные ряды на трейн и тест
#    len_seq- количество времменых точек в трейн
#    points_ahead - количество времменых точек в прогнозе
#    gap - расстояние между концом трейна и началом теста
#    intersection - если нет, то в выборке нет перескающих множеств (временнызх моментов)
#    shag - через сколько прыгаем
#    train_size - float от 0 до 1
#    
#    return list of dfs
#    
#    """
    #TODO требования к входным данным прописать
    #TODO переписать энергоэффективно чтобы было
    #TODO пока временные характеристики int_ами пора бы в pd.TimdDelta
    # нет индексов 
    assert len_seq + points_ahead + gap + 1 <= len(df)
    how='seq to seq'   

# -------------------------------------------------------  
#             
# -------------------------------------------------------  


    x_start=0
    x_end= x_start + len_seq
    y_start = x_end + gap +1
    y_end = y_start + points_ahead
    
    if intersection:
        # ради вычислительной нагрузки такой кастыль
        def compute_new_x_start(x_start,y_end,shag):
            return x_start + shag
    else:
        def compute_new_x_start(x_start,y_end,shag):
            return y_end + shag
    
    X = []
    y = []
    while y_end <= len(df):
        X.append(df[x_start:x_end])
        y.append(df[y_start:y_end])
        
        x_start= compute_new_x_start(x_start,y_end,shag)
        x_end= x_start + len_seq
        y_start = x_end + gap +1
        y_end = y_start + points_ahead
          
    
    if (test_size==0) | (len(X)==1):
        indices = np.array(range(len(X)))
        #             np.random.seed(random_state)
        if shuffle:
            print(indices)
            np.random.shuffle(indices)
            print(indices)
        X = [X[i] for i in indices]
        y = [y[i] for i in indices]
        return X,[],y,[]
    else:
        return train_test_split(X,y, 
                                test_size=test_size, 
                                train_size=train_size, 
                                random_state=random_state, 
                                shuffle=shuffle, 
                                stratify=stratify
                               )
						   
						   
						   
class Loader:
    def __init__(self, X,y, batch_size,shuffle=True):
        if shuffle==True:
            indices = np.array(range(len(X)))
            np.random.shuffle(indices)
            self.X = X[indices]
            self.y = y[indices]
        else:
            self.X = X
            self.y = y
        self.batch_size = batch_size

    def __iter__(self):
        self.i = - self.batch_size
        return self

    def __next__(self):
        if self.i+self.batch_size < len(self.X):
            self.i+=self.batch_size
            return self.X[self.i:self.i+self.batch_size], self.y[self.i:self.i+self.batch_size]
        elif self.i+2*self.batch_size < len(self.X):        
            return self.X[self.i:], self.y[self.i:]
        else:
            raise StopIteration
            
    def __len__(self):
        return len(np.arange(0,len(self.X),self.batch_size))
		        
        
def split_by_repeated(series,df=None):
    """
    retrun dict with lists of ts whwre keys is unique values
    """
    series = series.copy().dropna()
    if len(series.unique())==1:
        result = {series.unique()[0]:series}
    elif len(series.unique())>1:
        result = {uni:[] for uni in series.unique()}
        recent_i=0
        recent_val=series.values[0]
        for i in range(len(series)):
            val = series.values[i]
            if (recent_val == val):
                continue
            else:
                result[recent_val].append(series[recent_i:i])
                recent_i=i
                recent_val = val

        if i == len(series)-1:
            if (recent_val == val):
                result[recent_val].append(series[recent_i:i+1])
            else:
                result[recent_val].append(series[recent_i:i+1])
    else:
        raise NameError('0 series')


    if df is not None:
        new_result = {uni:[] for uni in series.unique()}
        for key in result:
            for i in range(len(result[key])):
                if len(result[key][i]) <=1:
                    continue
                else:
                    new_result[key].append(df.loc[result[key][i].index])
        return new_result
    else:
        return result
        

def evaluating_change_point(true, prediction, metric='nab', 
                            numenta_time=None,
                            portion=0.1,
                            anomaly_window_destenation='left',
                            intersection_mode='cut left',
                            table_of_coef=None,
                            scale_func = "default",
                            scale_koef=1,
                            hidden_anomalies_mode = True,
                            plot_figure=False,
                            change_point_mode = True,
                            verbose=True
                            ):
    """
    
    true - both:
                list of pandas Series with binary int labels
                pandas Series with binary int labels
    prediction - both:
                      list of pandas Series with binary int labels
                      pandas Series with binary int labels
    metric: 'nab', 'binary' (FAR, MAR), 'average_delay'. Default='nab'
    
    verbose:  : booleant, default=True.
                  Если True, то вывод полезной информации в output
    
    <<< For 'nab', 'average_delay' metrics: ...
        portion : float, default=0.1
                Параметр нужен, если numenta_time=None
                Ширина scoring window в этом случае равна portion от 
                ширины всей выборки деленной на количество real CPs в 
                этой выборке
        anomaly_window_destenation: 'left', 'right', 'center'. Defualt='right'
                Параметр размещения scoring window относительно аномалии. 
                'left'  : scoring window будет по левую сторону от аномалии
                'right' : scoring window будет по правую сторону от аномалии
                'center': scoring window будет по размещено относительно центра
                          аномалии

        intersection_mode: 'cut left', 'cut right', 'both'. Defualt='right'
                Параметр будет задействован, в случае, если scoring windows
                пересекаются, что в общем случае нежалательно, и желательно
                требует иного подхода, чем просто обрезки scoring window
                c помощью данного параметра.
                'cut left'  : обрежет пересакающуюся часть левого окна
                'cut right' : обрежет пересакающуюся часть правого окна
                'both'  : обрежет пересакающуюся часть и левого и правого окна

    ... For 'nab', 'average_delay' metrics>>> 
    <<< For 'nab' metric:
        table_of_coef: pd.DataFrame of specific form. See bellow. 
                Application profiles of NAB metric.If Defaut is None:  
                table_of_coef = pd.DataFrame([[1.0,-0.11,1.0,-1.0],
                                              [1.0,-0.22,1.0,-1.0],
                                              [1.0,-0.11,1.0,-2.0]])
                table_of_coef.index = ['Standart','LowFP','LowFN']
                table_of_coef.index.name = "Metric"
                table_of_coef.columns = ['A_tp','A_fp','A_tn','A_fn']
            
    scale_func : "default" of "improved". Defualt="default".
                 Scoring function in NAB metric.
                 'default'  : standart NAB scoring function 
                 'improved' : Our function for resolving disadvantages
                              of standart NAB scoring function 
                  
    scale_koef : float > 0. Defualt=1.0. 
                 Коэффициент сглаживания. Чем меньше он, тем более
                 сглажена scoring function
                
    hidden_anomalies_mode : boolean, default=True.
                            True : тогда слева значение Scoring function равно Atp, а справа Afp
                            False: тогда слева значение Scoring function равно Afp, а справа Atp
    
    plot_figure : booleant, default=False.
                  Если True, то прорисовка score fuction, например, 
                  чтобы откалибровать scale_koef
    
    change_point_mode : boolean, default=True.
                        Если False, тогда берется не только первая точка прогноза в 
                        scoring window, а все. При этом взешиваются, суммируются и нормируется так,
                        что по сути можно просто смотреть на рез-тат. 
    
    
                
    """
    
    def binary(true, prediction):      
        """
        true - true binary series with 1 as anomalies
        prediction - trupredicted binary series with 1 as anomalies
        """
        def single_binary(true,prediction):
            true_ = true == 1 
            prediction_ = prediction == 1
            TP = (true_ & prediction_).sum()
            TN = (~true_ & ~prediction_).sum()
            FP = (~true_ & prediction_).sum()
            FN = (true_ & ~prediction_).sum()
            return TP,TN,FP,FN
            
        if type(true) != type(list()):
            TP,TN,FP,FN = single_binary(true,prediction)
        else:
            TP,TN,FP,FN = 0,0,0,0
            for i in range(len(true)):
                TP_,TN_,FP_,FN_ = single_binary(true[i],prediction[i])
                TP,TN,FP,FN = TP+TP_,TN+TN_,FP+FP_,FN+FN_       
    
        f1 = round(TP/(TP+(FN+FP)/2), 2)
        print(f'False Alarm Rate {round(FP/(FP+TN)*100,2)} %' )
        print(f'Missing Alarm Rate {round(FN/(FN+TP)*100,2)} %')
        print(f'F1 metric {f1}')
        return f1
    
    def average_delay(detecting_boundaries, prediction):
        """
        Возвращает add, missing, all_true_anom
        
        detecting_boundaries - где правая граница трушная аномалия, только так
        """
        
        def single_average_delay(detecting_boundaries, prediction):
            missing = 0
            detectHistory = []
            all_true_anom = 0
            _df_fill_bounds =  pd.DataFrame(np.ones((len(prediction),len(detecting_boundaries)))*np.nan,index=prediction.index)
            for i in range(len(detecting_boundaries)):
                t1 = detecting_boundaries[i][0]
                t2 = detecting_boundaries[i][1]
                _df_fill_bounds.loc[t1:t2,i]=1
                if prediction[t1:t2].sum()==0:
                    missing+=1
                else:
                    detectHistory.append(prediction[prediction ==1][t1:t2].index[0]-t1)
                all_true_anom+=1
            # FPs
                        
            
                
            ts_fp = pd.Series(np.ones(len(prediction)),index=prediction.index)
            ts_fp.loc[_df_fill_bounds.dropna(how='all').index]=0
            ts_fp = ts_fp * prediction
            FP = ts_fp.sum()
            return missing, detectHistory, FP, all_true_anom
            
        
        if type(prediction) != type(list()):
            missing, detectHistory, FP, all_true_anom = single_average_delay(detecting_boundaries, prediction)
        else:
            missing, detectHistory, FP, all_true_anom = 0, [], 0, 0
            for i in range(len(prediction)):
                missing_, detectHistory_, FP_, all_true_anom_ = single_average_delay(detecting_boundaries[i], prediction[i])
                missing, detectHistory, FP, all_true_anom = missing+missing_, detectHistory+detectHistory_, FP+FP_, all_true_anom+all_true_anom_

        add = pd.Series(detectHistory).mean()
        if verbose:
            print('Amount of true anomalies',all_true_anom)
            print(f'A number of missed CPs = {missing}')
            print(f'A number of FPs = {FP}')
            print('Average delay', add)
            
        return add, missing, int(FP), all_true_anom
    
    def evaluate_nab(detecting_boundaries, prediction, 
                     table_of_coef=None,
                     intersection_mode='cut left',
                     hidden_anomalies_mode = False,
                     scale_func = "default",
                     scale_koef=1,
                     plot_figure=True,
                     change_point_mode = False
                    ):
        """
        Scoring labeled time series by means of
        Numenta Anomaly Benchmark methodics
        Parameters
        ----------
        detecting_boundaries: list of list of two float values
            The list of lists of left and right boundary indices
            for scoring results of labeling
        prediction: pd.Series with timestamp indices, in which 1 
            is change point, and 0 in other case. 
        table_of_coef: pandas array (3x4) of float values
            Table of coefficients for NAB score function
            indeces: 'Standart','LowFP','LowFN'
            columns:'A_tp','A_fp','A_tn','A_fn'
        Returns
        -------
        Scores: numpy array, shape of 3, float
            Score for 'Standart','LowFP','LowFN' profile 
        Scores_null: numpy array, shape 3, float
            Null score for 'Standart','LowFP','LowFN' profile             
        Scores_perfect: numpy array, shape 3, float
            Perfect Score for 'Standart','LowFP','LowFN' profile  
        """
        def single_evaluate_nab(detecting_boundaries,
                                prediction, 
                                table_of_coef=None,
                                intersection_mode='cut left',
                                hidden_anomalies_mode = False,
                                scale_func = "default",
                                scale_koef=1,
                                plot_figure=True,
                                change_point_mode = False
                               ):
            """
            недостаттки scale_func default  -
            1 - зависит от относительного шага, а это значит, что если 
            слишком много точек в scoring window то перепад будет слишком
            жестким в середение. 
            2-   то самая левая точка не равно  Atp, а права не равна Afp
            (особенно если пррименять расплывающую множитель)
   
            hidden_anomalies_mode тогда слева от границы Atp срправа Afp,
            иначе fault mode, когда слева от границы Afp срправа Atp
            """
      
            def sigm_scale(len_ts, A_tp, A_fp, koef=1):
                x = np.arange(-int(len_ts/2), len_ts - int(len_ts/2))
                
                x = x if hidden_anomalies_mode else x[::-1]
                y = (A_tp-A_fp)*(1/(1+np.exp(5*x*koef))) + A_fp
                return y
            def my_scale(len_ts,A_tp,A_fp,koef=1):
                """ts - участок на котором надо жахнуть окно """
                x = np.linspace(-np.pi/2,np.pi/2,len_ts)
                x = x if hidden_anomalies_mode else x[::-1]
                # Приведение если неравномерный шаг.
                #x_new = x_old * ( np.pi / (x_old[-1]-x_old[0])) - x_old[0]*( np.pi / (x_old[-1]-x_old[0])) - np.pi/2
                y = (A_tp-A_fp)/2*-1*np.tanh(koef*x)/(np.tanh(np.pi*koef/2)) + (A_tp-A_fp)/2 + A_fp
                return y 
            
            if scale_func == "default":
                scale_func = sigm_scale
            elif scale_func == "improved":
                scale_func = my_scale
            else:
                raise("choose the scale_func")
            
            if table_of_coef is None:
                table_of_coef = pd.DataFrame([[1.0,-0.11,1.0,-1.0],
                                     [1.0,-0.22,1.0,-1.0],
                                      [1.0,-0.11,1.0,-2.0]])
                table_of_coef.index = ['Standart','LowFP','LowFN']
                table_of_coef.index.name = "Metric"
                table_of_coef.columns = ['A_tp','A_fp','A_tn','A_fn']

            detecting_boundaries = detecting_boundaries.copy()
            prediction = prediction.copy()
            _df_fill_bounds =  pd.DataFrame(np.ones((len(prediction),len(detecting_boundaries)))*np.nan,index=prediction.index)
            for i in range(len(detecting_boundaries)):
                _df_fill_bounds.loc[detecting_boundaries[i][0]:detecting_boundaries[i][1],i]=1
                if plot_figure:
                    plt.figure()
                    ts = _df_fill_bounds.iloc[:,i].dropna()
                    ts_profile = pd.Series(data=scale_func( len(ts),
                                                            table_of_coef['A_tp']['Standart'],
                                                            table_of_coef['A_fp']['Standart'],
                                                            koef=scale_koef),
                                           index = ts.index)
                    plt.plot(ts_profile,label='profile')
                    ind_ = prediction[prediction==1].loc[ts.index[0]:ts.index[-1]].index
                    if len(ind_)>0:
                        for pred_anom in ind_:
                            plt.axvline(pred_anom,c='r',label='predicted anomaly')
                        plt.legend()
                        plt.title(str(i))
                        plt.show()
                
            
            Scores, Scores_perfect, Scores_null=[], [], []
            for profile in ['Standart', 'LowFP', 'LowFN']:       
                A_tp = table_of_coef['A_tp'][profile]
                A_fp = table_of_coef['A_fp'][profile]
                A_fn = table_of_coef['A_fn'][profile]
                
                score = 0
                # FPs
                ts_fp = pd.Series(np.ones(len(prediction)),index=prediction.index)
                ts_fp.loc[_df_fill_bounds.dropna(how='all').index]=0
                ts_fp = ts_fp * prediction
                score += A_fp*ts_fp.sum()
                #FNs and TPs
                for i in range(len(detecting_boundaries)):
                    ts_tp = _df_fill_bounds.iloc[:,i].dropna()
                    ts_tp = ts_tp * prediction.loc[ts_tp.index]
                    if ts_tp.sum()==0:
                        score+=A_fn
                    else:
                        ts_profile = pd.Series(data=scale_func(
                                                                len(ts_tp),A_tp,A_fp,koef=scale_koef),
                                               index = ts_tp.index)
                        if change_point_mode:
                            ts_tp.loc[ts_tp[ts_tp==1].index[1:]] = 0
                            score += (ts_profile * ts_tp).sum()
                        else:
                            score += (ts_profile*ts_tp).sum()/(ts_profile).sum()*len(ts_profile)
                Scores.append(score)
                Scores_perfect.append(len(detecting_boundaries)*A_tp)
                Scores_null.append(len(detecting_boundaries)*A_fn)
            return np.array([np.array(Scores),np.array(Scores_null), np.array(Scores_perfect)])
       #======      
        if type(prediction) != type(list()):
            matrix = single_evaluate_nab(detecting_boundaries, 
                                         prediction, 
                                         table_of_coef=table_of_coef,
                                         intersection_mode=intersection_mode,
                                         hidden_anomalies_mode = hidden_anomalies_mode,
                                         scale_func = scale_func,
                                         scale_koef=scale_koef,
                                         plot_figure=plot_figure,
                                         change_point_mode = change_point_mode)
        else:
            matrix = np.zeros((3,3))
            for i in range(len(prediction)):
                matrix_ = single_evaluate_nab(detecting_boundaries[i],
                                              prediction[i],
                                              table_of_coef=table_of_coef,
                                              intersection_mode=intersection_mode,
                                              hidden_anomalies_mode = hidden_anomalies_mode,
                                              scale_func = scale_func,
                                              scale_koef=scale_koef,
                                              plot_figure=plot_figure,
                                              change_point_mode = change_point_mode)
                matrix = matrix + matrix_      
                
        results = {}
        desc = ['Standart', 'LowFP', 'LowFN'] 
        for t, profile_name in enumerate(desc):
            results[profile_name] = round(100*(matrix[0,t]-matrix[1,t])/(matrix[2,t]-matrix[1,t]), 2)
            print(profile_name,' - ', results[profile_name])
        
        return results
            
            
    #=========================================================================
    if type(true) != type(list()):
        true_1_indexes = true[true==1].index
    else:
        true_1_indexes = [true[i][true[i]==1].index for i in range(len(true))]
        

    if not metric=='binary':
        if metric=='average_delay':
            anomaly_window_destenation= 'left'
        def single_detecting_boundaries(true, prediction, numenta_time, true_1_indexes,
                                        anomaly_window_destenation='left',portion=0.1):
            detecting_boundaries=[]
            td = pd.Timedelta(numenta_time) if numenta_time is not None else \
                        pd.Timedelta((true.index[-1]-true.index[0])/len(true_1_indexes)*portion)  
            for val in true_1_indexes:
                if anomaly_window_destenation == 'left':
                    detecting_boundaries.append([val, val + td])
                elif anomaly_window_destenation == 'right':
                    detecting_boundaries.append([val - td, val])
                elif anomaly_destenation == 'center':
                    anomaly_window_destenation.append([val - td/2, val + td/2])
                else:
                    raise('choose anomaly_window_destenation')
                    
                    
            # block for resolving intersection problem:
            # важно не ошибиться, и всегда следить, чтобы везде правая граница далее
            # не включалась, иначе будет пересечения окон             
            new_detecting_boundaries = detecting_boundaries.copy()
            if new_detecting_boundaries[0][0] < prediction.index[0]:
                new_detecting_boundaries[0][0] = prediction.index[0]
            if new_detecting_boundaries[-1][-1] > prediction.index[-1]:
                new_detecting_boundaries[-1][-1] = prediction.index[-1]
            for i in range(len(new_detecting_boundaries)-1):
                if new_detecting_boundaries[i][1] >= new_detecting_boundaries[i+1][0]:
                    print(f'Intersection of scoring windows{new_detecting_boundaries[i][1], new_detecting_boundaries[i+1][0]}')
                    if intersection_mode == 'cut left':
                        new_detecting_boundaries[i][1] = new_detecting_boundaries[i+1][0]
                    elif intersection_mode == 'cut right':
                        new_detecting_boundaries[i+1][0] = new_detecting_boundaries[i][1]
                    elif intersection_mode == 'cut both':
                        _a  = new_detecting_boundaries[i][1]
                        new_detecting_boundaries[i][1] = new_detecting_boundaries[i+1][0]
                        new_detecting_boundaries[i+1][0] = _a
                    else:
                        raise("choose the intersection_mode")
            detecting_boundaries = new_detecting_boundaries.copy()        
                    
                    
                    
            
            return detecting_boundaries
        
        if type(true) != type(list()):
            detecting_boundaries = single_detecting_boundaries(true=true,
                                                               prediction=prediction, 
                                                               numenta_time=numenta_time, 
                                                               true_1_indexes=true_1_indexes,
                                                               anomaly_window_destenation=anomaly_window_destenation,
                                                               portion=portion
                                                              )
        else:
            detecting_boundaries=[]
            for i in range(len(true)):
                detecting_boundaries.append(single_detecting_boundaries(true=true[i], 
                                                                        prediction=prediction[i],                                                                
                                                                        numenta_time=numenta_time, 
                                                                        true_1_indexes=true_1_indexes[i],
                                                                        anomaly_window_destenation=anomaly_window_destenation,
                                                                        portion=portion))


    if metric== 'nab':
        return evaluate_nab(detecting_boundaries, 
                            prediction,
                            table_of_coef=table_of_coef,
                            intersection_mode=intersection_mode,
                            hidden_anomalies_mode = hidden_anomalies_mode,
                            scale_func = scale_func,
                            scale_koef=scale_koef,
                            plot_figure=plot_figure,
                            change_point_mode = change_point_mode)
    elif metric=='average_delay':
        return average_delay(detecting_boundaries, prediction)
    elif metric== 'binary':
        return binary(true, prediction)
        
        
def df2dfs(df,  # Авторы не рекомендуют так делать,
            resample_freq = None, # требования
            thereshold_gap = None, 
            koef_freq_of_gap = 1.2, # 1.2 проблема которая возникает которую 02.09.2021 я написал в ИИ 
            plot = True,
            col = None):
    """
    Функция которая преообратает raw df до требований к входу на DL_AD:
    
    то есть разбивает df на лист of dfs by gaps 
    
    Не ресемлирует так как это тяжелая задача, но если частота реже чем 
    koef_freq_of_gap of thereshold_gap то это воспринимается как пропуск. 
    Основной посыл: если сигнал приходит чаще, то он не уползает сильно, 
    а значит не приводит к аномалии, а если редко то приводит, поэтому воспри-
    нимается как пропуск. 
    
    Parameters
    ----------   
    df : pd.DataFrame
        Исходный временной ряд полностью за всю историю   
        
    resample_freq: pd.TimeDelta (optional, default=None)
        Частота дискретизации временного ряда. 
        Если default то самая частая частота дискретизации. При этом, 
        если нет выраженной частоты вылетит ошибка. 
    thereshold_gap : pd.TimeDelta (optional, default=None)
        Порог периода, превышая который функция будет воспринимать
        данный период как пропуск. 
    koef_freq_of_gap : float or int (optional if thereshold_gap==None,
        default=1.2)  
        thereshold_gap = koef_freq_of_gap * resample_freq
    plot : bool (optional, default=True)
        If true, then отрисуется нарезка
        If false, then не отрисуется нарезка   
    col : int of str (optional, default=True)
        Название или номер колонки для отрисовки
        Если None первая колонка

    Returns
    -------
    dfs : list of pd.DataFrame
        Список времменных рядов без пропусков с относительно стабильной 
        частотой дискретизации. 
    """
    
    df = df.dropna(how='all').dropna(1,how='all')
    dts  = df.dropna(how='all').index.to_series().diff()
    if resample_freq is None:
        dts_dist = dts.value_counts()
        if dts_dist[0] > dts_dist[1:].sum():
            resample_freq  = dts_dist.index[0]
        else: 
            print(dts_dist)
            raise Exception("Необходимо самостоятельно обработать функцию так как нет преобладающей частоты дискретизации")
    thereshold_gap = resample_freq*koef_freq_of_gap if thereshold_gap is None else thereshold_gap
    gaps = (dts > thereshold_gap).astype(int).cumsum()
    dfs = [df.loc[gaps[gaps==stage].index] for stage in gaps.unique()]

    if plot:
        f, ax = plt.subplots()
        if col is None:
            col = df.columns[0]
        else:
            if type(col)==type(int):
                col = df.columns[col]
        for df in dfs:
            df[col].plot(ax=ax)
    return dfs
    
    


from torch import nn, optim
import torch
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
import numpy as np


class SimpleLSTM(nn.Module):
    def __init__(self, in_features, n_hidden, n_layers=2, bidirectional=False, dropout=0.5):
        super(SimpleLSTM, self).__init__()
        self.n_hidden = n_hidden
        self.n_layers = n_layers
        self.lstm = nn.LSTM(input_size=in_features,
                            hidden_size=n_hidden,
                            num_layers=n_layers,
                            dropout=dropout,
                            batch_first =True,
                            bidirectional = bidirectional
                           )
        self.k_bidir = 2 if bidirectional else 1 

        self.linear = nn.Linear(in_features=n_hidden, out_features=in_features)
    
    
    def initHidden(self,batch_size):
        self.hidden = (
            torch.zeros(self.n_layers*self.k_bidir, batch_size, self.n_hidden).to(device),
            torch.zeros(self.n_layers*self.k_bidir, batch_size, self.n_hidden).to(device)
        )
    def forward(self, sequences):
        batch_size  = len(sequences)
        lstm_out, self.hidden = self.lstm(sequences, self.hidden )
        last_time_step = lstm_out.reshape(-1, batch_size, self.n_hidden)[-1] # -1 is len_seq

        y_pred = self.linear(last_time_step)
        return y_pred

    def run_epoch(self, iterator, optimizer, criterion, points_ahead=1, phase='train',encod_decode_model=False):
        is_train = (phase == 'train')
        if is_train:
            self.train()
        else:
            self.eval()
        
        epoch_loss = 0
        
        if points_ahead !=1:
            assert (points_ahead > 0) & (type(points_ahead)==type(int()))
            def forecast_multistep(y_pred,points_ahead):
                new_x = y_pred
                for j in range(points_ahead-1):
                    new_x = self.forward(new_x).unsqueeze(1)
                    y_pred = torch.cat([y_pred,new_x],1)
                return y_pred
        else:
            def forecast_multistep(y_pred,points_ahead):
                return y_pred

        all_y_preds = []
        with torch.set_grad_enabled(is_train):
            for i, (x,y) in enumerate(iterator):
                x,y = np.array(x),np.array(y) #df.index rif of
                self.initHidden(x.shape[0])
                
                x = torch.tensor(x).float().to(device).requires_grad_()
                y_true = torch.tensor(y).float().to(device)
                y_pred = self.forward(x).unsqueeze(1)
                y_pred = forecast_multistep(y_pred,points_ahead)
                
                if encod_decode_model:
                    y_pred = torch.cat([y_pred[:,i,:].unsqueeze(1) for i in range(y_pred.shape[1]-1,-1,-1)],1)
                
                if phase == 'forecast':
                    all_y_preds.append(y_pred)
                    continue # in case of pahse = 'forecast' criterion is None
                        
                loss = criterion(y_pred,y_true)
                if is_train:
                  optimizer.zero_grad()
                  loss.backward()
                  optimizer.step()

                
                epoch_loss += loss.item()
        if phase != 'forecast':
            return epoch_loss / len(iterator)#, n_true_predicted / n_predicted
        else:
            return torch.cat(all_y_preds).detach().cpu().numpy()






####################
#=======================================================
#              From https://github.com/Zhang-Zhi-Jie/Pytorch-MSCRED/blob/master/utils/matrix_generator.py
#=======================================================







#=======================================================
#              From other sourse
#=======================================================
# class ConvLSTMCell(nn.Module):

#     def __init__(self, input_dim, hidden_dim, kernel_size, bias):
#         """
#         MIT license
#         Authors and any other relevant information you can see:  
#         https://github.com/ndrplz/ConvLSTM_pytorch 
        
        
#         Initialize ConvLSTM cell.
#         Parameters
#         ----------
#         input_dim: int
#             Number of channels of input tensor.
#         hidden_dim: int
#             Number of channels of hidden state.
#         kernel_size: (int, int)
#             Size of the convolutional kernel.
#         bias: bool
#             Whether or not to add the bias.
#         """

#         super(ConvLSTMCell, self).__init__()

#         self.input_dim = input_dim
#         self.hidden_dim = hidden_dim

#         self.kernel_size = kernel_size
#         self.padding = kernel_size[0] // 2, kernel_size[1] // 2
#         self.bias = bias

#         self.conv = nn.Conv2d(in_channels=self.input_dim + self.hidden_dim,
#                               out_channels=4 * self.hidden_dim,
#                               kernel_size=self.kernel_size,
#                               padding=self.padding,
#                               bias=self.bias)

#     def forward(self, input_tensor, cur_state):
#         h_cur, c_cur = cur_state

#         combined = torch.cat([input_tensor, h_cur], dim=1)  # concatenate along channel axis

#         combined_conv = self.conv(combined)
#         cc_i, cc_f, cc_o, cc_g = torch.split(combined_conv, self.hidden_dim, dim=1)
#         i = torch.sigmoid(cc_i)
#         f = torch.sigmoid(cc_f)
#         o = torch.sigmoid(cc_o)
#         g = torch.tanh(cc_g)

#         c_next = f * c_cur + i * g
#         h_next = o * torch.tanh(c_next)

#         return h_next, c_next

#     def init_hidden(self, batch_size, image_size):
#         height, width = image_size
#         return (torch.zeros(batch_size, self.hidden_dim, height, width, device=self.conv.weight.device),
#                 torch.zeros(batch_size, self.hidden_dim, height, width, device=self.conv.weight.device))


# class ConvLSTM(nn.Module):

#     """
#     MIT license
#     Authors and any other relevant information you can see:  
#     https://github.com/ndrplz/ConvLSTM_pytorch 
    
    
    
#     Parameters:
#         input_dim: Number of channels in input
#         hidden_dim: Number of hidden channels
#         kernel_size: Size of kernel in convolutions
#         num_layers: Number of LSTM layers stacked on each other
#         batch_first: Whether or not dimension 0 is the batch or not
#         bias: Bias or no bias in Convolution
#         return_all_layers: Return the list of computations for all layers
#         Note: Will do same padding.
#     Input:
#         A tensor of size B, T, C, H, W or T, B, C, H, W
#     Output:
#         A tuple of two lists of length num_layers (or length 1 if return_all_layers is False).
#             0 - layer_output_list is the list of lists of length T of each output
#             1 - last_state_list is the list of last states
#                     each element of the list is a tuple (h, c) for hidden state and memory
#     Example:
#         >> x = torch.rand((32, 10, 64, 128, 128))
#         >> convlstm = ConvLSTM(64, 16, 3, 1, True, True, False)
#         >> _, last_states = convlstm(x)
#         >> h = last_states[0][0]  # 0 for layer index, 0 for h index
#     """

#     def __init__(self, input_dim, hidden_dim, kernel_size, num_layers,
#                  batch_first=False, bias=True, return_all_layers=False):
#         super(ConvLSTM, self).__init__()

#         self._check_kernel_size_consistency(kernel_size)

#         # Make sure that both `kernel_size` and `hidden_dim` are lists having len == num_layers
#         kernel_size = self._extend_for_multilayer(kernel_size, num_layers)
#         hidden_dim = self._extend_for_multilayer(hidden_dim, num_layers)
#         if not len(kernel_size) == len(hidden_dim) == num_layers:
#             raise ValueError('Inconsistent list length.')

#         self.input_dim = input_dim
#         self.hidden_dim = hidden_dim
#         self.kernel_size = kernel_size
#         self.num_layers = num_layers
#         self.batch_first = batch_first
#         self.bias = bias
#         self.return_all_layers = return_all_layers

#         cell_list = []
#         for i in range(0, self.num_layers):
#             cur_input_dim = self.input_dim if i == 0 else self.hidden_dim[i - 1]

#             cell_list.append(ConvLSTMCell(input_dim=cur_input_dim,
#                                           hidden_dim=self.hidden_dim[i],
#                                           kernel_size=self.kernel_size[i],
#                                           bias=self.bias))

#         self.cell_list = nn.ModuleList(cell_list)

#     def forward(self, input_tensor, hidden_state=None):
#         """
#         Parameters
#         ----------
#         input_tensor: todo
#             5-D Tensor either of shape (t, b, c, h, w) or (b, t, c, h, w)
#         hidden_state: todo
#             None. todo implement stateful
#         Returns
#         -------
#         last_state_list, layer_output
#         """
#         if not self.batch_first:
#             # (t, b, c, h, w) -> (b, t, c, h, w)
#             input_tensor = input_tensor.permute(1, 0, 2, 3, 4)

#         b, _, _, h, w = input_tensor.size()

#         # Implement stateful ConvLSTM
#         if hidden_state is not None:
#             raise NotImplementedError()
#         else:
#             # Since the init is done in forward. Can send image size here
#             hidden_state = self._init_hidden(batch_size=b,
#                                              image_size=(h, w))

#         layer_output_list = []
#         last_state_list = []

#         seq_len = input_tensor.size(1)
#         cur_layer_input = input_tensor

#         for layer_idx in range(self.num_layers):

#             h, c = hidden_state[layer_idx]
#             output_inner = []
#             for t in range(seq_len):
#                 h, c = self.cell_list[layer_idx](input_tensor=cur_layer_input[:, t, :, :, :],
#                                                  cur_state=[h, c])
#                 output_inner.append(h)

#             layer_output = torch.stack(output_inner, dim=1)
#             cur_layer_input = layer_output

#             layer_output_list.append(layer_output)
#             last_state_list.append([h, c])

#         if not self.return_all_layers:
#             layer_output_list = layer_output_list[-1:]
#             last_state_list = last_state_list[-1:]

#         return layer_output_list, last_state_list

#     def _init_hidden(self, batch_size, image_size):
#         init_states = []
#         for i in range(self.num_layers):
#             init_states.append(self.cell_list[i].init_hidden(batch_size, image_size))
#         return init_states

#     @staticmethod
#     def _check_kernel_size_consistency(kernel_size):
#         if not (isinstance(kernel_size, tuple) or
#                 (isinstance(kernel_size, list) and all([isinstance(elem, tuple) for elem in kernel_size]))):
#             raise ValueError('`kernel_size` must be tuple or list of tuples')

#     @staticmethod
#     def _extend_for_multilayer(param, num_layers):
#         if not isinstance(param, list):
#             param = [param] * num_layers
#         return param


def feature_importance(residuals, analysis_type="collective", date_from=None, date_till=None, weigh=True):
    """Feature importance calculation

    Parameters
    ----------
    residuals : pandas.DataFrame()

    analysis_type : str, "single"/"collective", "single" by default

    date_from : str in format 'yyyy-mm-dd HH:MM:SS', None by default

    date_till : str in format 'yyyy-mm-dd HH:MM:SS', None by default

    weigh : boolean, True by default
        If analysis_type == "collective".

    Returns
    -------
    data : pandas.DataFrame().
    """
    if date_from is None:
        start = 0
    if date_till is None:
        end = -1
    data = residuals[date_from:date_till].abs().copy()

    if (analysis_type == "collective") & (weigh == False):
        data = data.div(data.sum(axis=1), axis=0) * 100
        return pd.DataFrame(data.mean(), columns=['Feature importance, %']).T
    elif (analysis_type == "collective") & (weigh == True):
        data = data.mean().div(data.mean().sum(), axis=0) * 100
        return pd.DataFrame(data, columns=['Feature importance, %']).T
    elif analysis_type == "single":
        return data.div(data.sum(axis=1), axis=0) * 100   



from sklearn.preprocessing import MinMaxScaler, StandardScaler
import numpy as np
import pandas as pd
import torch
from torch import nn
import matplotlib.pyplot as plt
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")



from . import src
from . import models
from . import generate_residuals
from . import stastics


class DL_AD():
    """        
    Пайплайн Time Series Anomaly Detection based on 
    SOTA deep learning forecasting algorithms.
    
    Данный пайплайн избавит вас от проблем написания кода для: \n
    1) формирование выборок для подачи в sequence модели \n
    2) обучения моделей \n
    3) поиска аномалий в невязках \n
    
    Данный пайплайн позволяет: \n
    1) пронгозировать временные ряды, в том числе многомерные. \n
    2) вычислять невязку между прогнозом и настоящими значениями \n
    3) анализировать невязку, и возращать разметку аномалиями \n
    
    Parameters
    ----------
    preproc : object, default = sklearn.preprocessing.MinMaxScaler()
        Объект предобратки значений временного ряда.
        Требования к классу по методами атрибутам одинаковы с default.
    
    generate_res_func : func, default = generate_residuals.abs
        Функция генерация невязки. На вход y_pred, y_true. В default это
        абсолютная разница значений. Требования к функциям описаны в 
        generate_residuals.py. 
        
    res_analys_alg : object, default=stastics.Hotelling().
        Объект поиска аномалий в остатках. В default это
        статистика Хоттелинга.Требования к классам описаны в 
        generate_residuals.py. 
        
    
    Attributes
    ----------
    
    
    Return 
    ----------
    object : object Объект этого класса DL_AD

    References
    ----------
    
    Links to the papers 

    Examples
    --------
    https://github.com/waico/tsad/tree/main/examples 
    """
    def _get_Train_Test_sets(self,dfs,len_seq,
                             points_ahead,
                             gap,
                             shag,
                             intersection,
                             test_size,
                             train_size,
                             random_state,
                             shuffle,  
                             stratify,
                            ):
        """
        Вспомогательная функция, избавляющая от дубляжа
        """

        if (type(dfs)== pd.core.series.Series) | (type(dfs) == pd.core.frame.DataFrame):
            df = dfs.copy() if type(dfs) == pd.core.frame.DataFrame else pd.DataFrame(dfs)
            self.columns = df.columns
            self.index = df.index
            if self._init_preproc:
                new_df = pd.DataFrame(self.preproc.fit_transform(df),index=df.index,columns=df.columns)
                self._init_preproc = False
            else:
                new_df = pd.DataFrame(self.preproc.transform(df),index=df.index,columns=df.columns)
            assert len_seq + points_ahead + gap -1 <= len(df)                       
            X_train, X_test, y_train, y_test= src.ts_train_test_split(df = new_df,
                                                                      len_seq=len_seq,
                                                                      points_ahead=points_ahead,
                                                                      gap=gap,
                                                                      shag=shag,
                                                                      intersection=intersection,
                                                                      test_size=test_size,
                                                                      train_size=train_size,
                                                                      random_state=random_state,
                                                                      shuffle=False, # потому что потом нужно в основном итераторе
                                                                      stratify=stratify)

        elif type(dfs) == type(list()):
            # уже все pd.DataFrame
            _df = pd.concat(dfs,ignore_index=True)
            if self._init_preproc:
                self.preproc.fit(_df)
                self._init_preproc = False            
            self.columns = _df.columns
            self.index  =  _df.index

            X_train, X_test, y_train, y_test = [],[],[],[]
            _k = 0
            for df in dfs:
                if ((type(df) == pd.core.series.Series) | (type(df) == pd.core.frame.DataFrame))==False:
                    raise NameError('Type of dfs is unsupported') 
                if not (len_seq + points_ahead + gap + 1 <= len(df)):
                     _k+=1
                     continue
                
                new_df = pd.DataFrame(self.preproc.transform(df),index=df.index,columns=df.columns)
                _X_train, _X_test, _y_train, _y_test= src.ts_train_test_split(new_df,len_seq,
                                                                          points_ahead=points_ahead,
                                                                          gap=gap,
                                                                          shag=shag,
                                                                          intersection=intersection,
                                                                          test_size=test_size,
                                                                          train_size=train_size,
                                                                          random_state=random_state,
                                                                          shuffle=False,
                                                                          stratify=stratify)
                X_train += _X_train
                X_test += _X_test
                y_train += _y_train
                y_test += _y_test
            
            print(f'Пропущено {_k} датастов, из-за того что saples слишком малов в датасете. (len_seq + points_ahead + gap -1 <= len(df))')

        else:
            raise NameError('Type of dfs is unsupported')  
            
        return X_train, X_test, y_train, y_test
    
    
    
    
    def _get_anomaly_timestamps(self,
                                dfs,
                                Loader,
                                len_seq,
                                points_ahead,
                                gap,
                                shag,
                                intersection,
                                test_size,  
                                random_state,
                                shuffle,
                                stratify,
                                point_ahead_for_residuals=0):
        """
        Вспомогательная функция для  генерации всего
        """
        X, _ , y_true, _  = self._get_Train_Test_sets(dfs=dfs,
                                                      len_seq=len_seq,
                                                      points_ahead=points_ahead, # 1 это default, так с остатками лучше не шутить до сих пор 
                                                      gap=gap,
                                                      shag=shag,
                                                      intersection=intersection,
                                                      test_size=test_size,  
                                                      train_size=None,
                                                      random_state=random_state,
                                                      shuffle=shuffle,
                                                      stratify=stratify)


        all_data_iterator = Loader(X,y_true, self.batch_size,shuffle=False)          
        y_pred = self.model.run_epoch(all_data_iterator, None, None, phase='forecast', points_ahead=points_ahead)     
        residuals = self.generate_res_func(y_pred,np.array(y_true))           
        point_ahead_for_residuals = 0 # мы иногда прогнозим на 10 точек вперед, ну интересует все равно на одну точку впреред 
        res_indices = [y_true[i].index[point_ahead_for_residuals] for i in range(len(y_true))]                                                
        df_residuals = pd.DataFrame(residuals[:,point_ahead_for_residuals,:],columns=self.columns,index=res_indices).sort_index()
        return df_residuals
        
        

        
# -----------------------------------------------------------------------------------------
#     Формирование сутевой части класса 
# -----------------------------------------------------------------------------------------    
    
    
    def __init__(self,     
                 preproc=None,
                 generate_res_func=None,
                 res_analys_alg=None,
                 
                ):
       
        
        self.preproc = MinMaxScaler() if preproc is None else preproc
        self.generate_res_func = generate_residuals.abs if generate_res_func is None else generate_res_func
        self.res_analys_alg = stastics.Hotelling() if res_analys_alg is None else res_analys_alg
        

        
    def fit(self,
            dfs,
            model=None,
            encod_decode_model = False,  # ужас, нужно это править, особенность encod_decode модели. Попытаться вообще еубрать эту переменную          
            criterion=None,
            optimiser=None,
            batch_size = 64,
            len_seq = 10,
            points_ahead = 5,
            n_epochs = 100,
            gap=0, 
            shag=1,
            intersection=True,
            test_size=0.2,
            train_size=None,
            random_state=None,
            shuffle=False,
            show_progress=True,
            show_figures=True,
            best_model_file = './best_model.pt',
            stratify=None,
            Loader=None,
            
           ):

        """
        Обучение модели как для задачи прогнозирования так и для задачи anomaly
        detection на имеющихся данных. fit = fit_predict_anmaloy 
        
        Parameters
        ----------
        dfs : {{df*,ts*}, list of {df*,ts*}}
            df*,ts* are pd.core.series.Seriesor or pd.core.frame.DataFrame data type.
            Исходные данные. Данные не долнжны содержать np.nan вовсе, иметь постоянную 
            и одинковую частоту of df.index и при этом не иметь пропусков. Проблему с 
            пропуском решают дробление одно df на list of df.             
        
        model : object of torch.nn.Module class, default=models.SimpleLSTM()
            Используемая модель нейронной сети. 
        
        criterion : object of torch.nn class, default=nn.MSELoss()
            Критерий подсчета ошибки для оптмизации. 
        
        optimiser : tuple = (torch.optim class ,default = torch.optim.Adam,
            dict  (dict of arguments without params models) , default=default)
            Example of optimiser : optimiser=(torch.optim.Adam,{'lr':0.001})
            Метод оптимизации нейронной сети и его параметры, указанные в 
            документации к torch.
            
        batch_size :  int, default=64
            Размер батча (Число сэмплов по которым усредняется градиент)
        
        len_seq : int, default=10
            Размер окна (количество последовательных точек ряда), на котором
            модель реально работает. По сути аналог порядка в авторегрессии. 
        
        points_ahead : int, default=5
            Горизонт прогнозирования. 
        
        n_epochs :  int, default=100 
            Количество эпох.
        
        >>> train_test_split vars
        
            gap :  int, default=0
                Сколько точек между трейном и тестом. Условно говоря,
                если крайняя точка train а это t, то первая точка теста t + gap +1.
                Параметр создан, чтобы можно было прогнозировать одну точку через большой 
                дополнительный интервал времени. 
            
            shag :  int, default=1.
                Шаг генерации выборки. Если первая точка была t у 1-ого сэмпла трейна,
                то у 2-ого сэмла трейна она будет t + shag, если intersection=True, иначе 
                тоже самое но без пересечений значений ряда. 
        
            intersection :  bool, default=True
                Наличие значений ряда (одного момента времени) в различных сэмплах выборки. 
            
            test_size : float or int, default=None
                If float, should be between 0.0 and 1.0 and represent the proportion
                of the dataset to include in the test split. If int, represents the
                absolute number of test samples. If None, the value is set to the
                complement of the train size. If ``train_size`` is also None, it will
                be set to 0.25. *
                *https://github.com/scikit-learn/scikit-learn/blob/95119c13a/sklearn/model_selection/_split.py#L2076 
                Может быть 0, тогда вернет значения X,y
            
            train_size : float or int, default=None
                If float, should be between 0.0 and 1.0 and represent the
                proportion of the dataset to include in the train split. If
                int, represents the absolute number of train samples. If None,
                the value is automatically set to the complement of the test size. *
                *https://github.com/scikit-learn/scikit-learn/blob/95119c13a/sklearn/model_selection/_split.py#L2076
            
            random_state : int, RandomState instance or None, default=None
                Controls the shuffling applied to the data before applying the split.
                Pass an int for reproducible output across multiple function calls.
                See :term:`Glossary <random_state>`.*
                *https://github.com/scikit-learn/scikit-learn/blob/95119c13a/sklearn/model_selection/_split.py#L2076
                
            
            shuffle : bool, default=True
                Whether or not to shuffle the data before splitting. If shuffle=False
                then stratify must be None. *
            
            show_progress : bool, default=True
                Показывать или нет прогресс обучения с детализацией по эпохам. 

            
            show_figures : bool, default=True
                Показывать или нет результаты решения задачии anomaly detection 
                и кривую трейна и валидации по эпохам. 
            
            
            best_model_file : string, './best_model.pt'
                Путь до файла, где будет хранится лучшие веса модели
            
            Loader : class, default=src.Loader.
                Тип загрузчика, которую будет использовать как итератор в будущем, 
                благодаря которому, есть возможность бить на бачи.
        
        Attributes
        ----------

        Return 
        ----------
        list of pd.datetime anomalies on initial dataset
        """
        self._init_preproc=True # это кастыль для _get_Train_Test_sets
        self.points_ahead = points_ahead
        self.len_seq = len_seq
        self.batch_size = batch_size
        dfs = dfs.copy()
        self.best_model_file = best_model_file
        self.encod_decode_model = encod_decode_model
        
         

# -----------------------------------------------------------------------------------------
#     Формирование train_iterator и val_iteraror
# -----------------------------------------------------------------------------------------
        if Loader is None:
            Loader = src.Loader
            
        
        X_train, X_test, y_train, y_test = self._get_Train_Test_sets(dfs=dfs,
                                                               len_seq=len_seq,
                                                               points_ahead=points_ahead,
                                                               gap=gap,
                                                               shag=shag,
                                                               intersection=intersection,
                                                               test_size=test_size,
                                                               train_size=train_size,
                                                               random_state=random_state,
                                                               shuffle=shuffle,  
                                                               stratify=stratify) 

        train_iterator = Loader(X_train, y_train,batch_size,shuffle=shuffle)
        val_iterator   = Loader(X_test, y_test,batch_size,shuffle=shuffle)

# -----------------------------------------------------------------------------------------
#     Обучение моделей
# -----------------------------------------------------------------------------------------
            

        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    
        
        if criterion is None:
            criterion = nn.MSELoss()
            
        if model is None:
            model = models.SimpleLSTM(len(self.columns),len(self.columns)).to(device)
        self.model = model.to(device)
            
            
        if optimiser is None:
            optimiser = torch.optim.Adam
            optimiser = optimiser(self.model.parameters())
        else:
            args = optimiser[1]
            args['params']=self.model.parameters()
            optimiser = optimiser[0](**args)
            
            
        history_train = []
        history_val = []
        best_val_loss = float('+inf')
        for epoch in range(n_epochs):    
            train_loss = self.model.run_epoch(train_iterator, optimiser, criterion, phase='train', points_ahead=points_ahead,encod_decode_model=self.encod_decode_model)#, writer=writer)
            val_loss = self.model.run_epoch(val_iterator, None, criterion, phase='val', points_ahead=points_ahead,encod_decode_model=self.encod_decode_model)#, writer=writer)

            history_train.append(train_loss)
            history_val.append(val_loss)

            if val_loss < best_val_loss:
                best_val_loss = val_loss
                torch.save(self.model.state_dict(), self.best_model_file)
            
            if show_progress:
                print(f'Epoch: {epoch+1:02}')
                print(f'\tTrain Loss: {train_loss:.3f}')
                print(f'\t Val. Loss: {val_loss:.3f} ') 

                
        self.model.load_state_dict(torch.load(self.best_model_file))
        
        if show_progress:
            try:
                test_iterator  = Loader(X_test, y_test, len(X_test),shuffle=False)
                test_loss = self.model.run_epoch(test_iterator, None, criterion, phase='val',encod_decode_model=self.encod_decode_model)
                print(f'Test Loss: {test_loss:.3f}')
            except:
                print('Весь X_test не помещается в память, тестим усреднением по батчам')
                test_iterator  = Loader(X_test, y_test,batch_size,shuffle=False)
                test_loss = []
                for epoch in range(n_epochs):  
                    test_loss.append(self.model.run_epoch(test_iterator, None, criterion, phase='val',encod_decode_model=self.encod_decode_model))
                print(f'Test Loss: {np.mean(test_loss):.3f}')
            
        if show_figures:
            plt.figure()
            plt.plot(history_train,label='Train')
            plt.plot(history_val,label='Val')
            plt.xlabel('Epoch')
            plt.ylabel('MSE')
            plt.legend()
            plt.show()

# -----------------------------------------------------------------------------------------
#     Генерация остатков
# -----------------------------------------------------------------------------------------            
        df_residuals =  self._get_anomaly_timestamps(dfs=dfs,
                                                    Loader=Loader,
                                                    len_seq=len_seq,
                                                    points_ahead=1,
                                                    gap=gap,
                                                    shag=shag,
                                                    intersection=intersection,
                                                    test_size=0,
                                                    random_state=None,
                                                    shuffle=False,  
                                                    stratify=stratify,
                                                    point_ahead_for_residuals=0)        
        self.anomaly_timestamps = self.res_analys_alg.fit_predict(df_residuals,show_figure=show_figures)
        self.statistic = self.res_analys_alg.statistic
        self.ucl = self.res_analys_alg.ucl
        self.lcl = self.res_analys_alg.lcl
        return self.anomaly_timestamps
        
       
    
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++   

   # накосячил тут с прогнозом на одну точку вперед. Могут быть проблемы если ahead !=1
    def predict_anomaly(self,
                        dfs,
                        Loader=None,
                        gap=0, 
                        shag=1,
                        intersection=True,
                        train_size=None,
                        random_state=None,
                        shuffle=False,
                        stratify=None,
                        show_progress=True,
                        show_figures=True
           ):
        
        
        """
        Поиск аномалий в новом наборе данных
        
        Parameters
        ----------
        см self.fit() dockstring
        
        
        Return
        ----------
        anomaly_timestamps : list of df.index.dtype
            Возвращает список временных меток аномалий                
        
        Attributes
        ----------
        
        """
        len_seq = self.len_seq
        

        if Loader is None:
            Loader = src.Loader


# -----------------------------------------------------------------------------------------
#     Генерация остатков
# -----------------------------------------------------------------------------------------            
        df_residuals =  self._get_anomaly_timestamps(dfs=dfs,
                                                    Loader=Loader,
                                                    len_seq=len_seq,
                                                    points_ahead=1,
                                                    gap=gap,
                                                    shag=shag,
                                                    intersection=intersection,
                                                    test_size=0,
                                                    random_state=None,
                                                    shuffle=False,  
                                                    stratify=stratify,
                                                    point_ahead_for_residuals=0)                                                                   
        self.anomaly_timestamps = self.res_analys_alg.predict(df_residuals,show_figure=show_figures)
        self.statistic = self.res_analys_alg.statistic
        return self.anomaly_timestamps


    
    def forecast(self,df,points_ahead=None,Loader=None, show_figures=True):
        """
        Прогнозирование временного ряда, в том числе векторного.
        
        Parameters
        ----------
        df : pd.core.series.Series or pd.core.frame.DataFrame data type
            Исходные данные. Данные не долнжны содержать np.nan вовсе, иметь постоянную 
            и одинковую частоту of df.index и при этом не иметь пропусков.         
                
        points_ahead : int, default=5
            Горизонт прогнозирования. 
               
        show_figures : bool, default=True
            Показывать или нет результаты решения задачии anomaly detection 
            и кривую трейна и валидации по эпохам. 
        
        
        Loader : class, default=src.Loader.
            Тип загрузчика, которую будет использовать как итератор в будущем, 
            благодаря которому, есть возможность бить на бачи.
        
                
        

        
        Attributes
        ----------
        
        """
        if Loader is None:
            Loader = src.Loader
        
        df  = df.copy()   
        points_ahead = points_ahead if points_ahead is not None else self.points_ahead
        len_seq = self.len_seq
        batch_size = self.batch_size         

        assert (type(df)== pd.core.series.Series) | (type(df) == pd.core.frame.DataFrame)
        df = df.copy() if type(df) == pd.core.frame.DataFrame else pd.DataFrame(df)
        df = df[-len_seq:]
        assert not self._init_preproc
        preproc_values = self.preproc.transform(df)
        
        
        iterator = Loader(np.expand_dims(preproc_values,0),np.expand_dims(preproc_values,0), #ничего страшного, 'y' все равно не используется
                          batch_size,shuffle=False)

        y_pred = self.model.run_epoch(iterator, None, None, phase='forecast', points_ahead=points_ahead)[0]
        y_pred = self.preproc.inverse_transform(y_pred)
        
        t_last = np.datetime64(df.index[-1])
        delta_dime = np.timedelta64(df.index[-1] - df.index[-2]) 
        new_index = pd.to_datetime(t_last + np.arange(1,points_ahead+1)*delta_dime)
        y_pred = pd.DataFrame(y_pred,index=new_index,columns=df.columns)
        
        if show_figures:
            pd.concat([df,y_pred])[-3*points_ahead:].plot()
            plt.axvspan(t_last,y_pred.index[-1], alpha=0.2, color='green',label='forecast')
            plt.xlabel('Datetime')
            plt.ylabel('Value')
            plt.legend()
            plt.show()
        
        return y_pred

    
 


# input y_pred,y_true of 3d, 

import numpy as np

def abs(y_pred,y_true):
    """
    Функция позволяющая получить разницу
    """
    assert y_pred.shape == y_true.shape
    return np.abs(y_pred-y_true)


#  Требования
#  работа как с одномерным pd.DataFrame так и с многомерными
# Наличие show_figure
# Наличие методов fit, predict, fit_predict
# Сохрание в атрибуты статистик и пределов: ucl, lcl, statistic



#
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd

"""
nan - недопустимы
"""
class Hotelling():
    def __init__(self,koef_ucl=3):
        self.koef_ucl = koef_ucl
    
    def fit(self,df):
        if df.shape[1]==1:
            self.inv_cov = np.array(1/ np.cov(df.T)).reshape(1,1)
        else:
            try: 
                self.inv_cov =  np.linalg.inv(np.cov(df.T))
            except:
                self.inv_cov =  np.linalg.pinv(np.cov(df.T))
        self.mean = df.mean()
        # try так как, когда много считает он в шоке. 
        try: 
            statistic  = (((df - self.mean).values @ self.inv_cov) @ (df - self.mean).values.T).diagonal()
        except:
            statistic = df.apply(lambda x: (((x - self.mean).values @ self.inv_cov) @ (x - self.mean).values.T) ,1 )
        
        self.ucl = statistic.mean()+self.koef_ucl*statistic.std()
        self.lcl = None
    
    def predict(self,df,show_figure=False):
        self.statistic  = pd.Series((((df - self.mean).values @ self.inv_cov) @ (df - self.mean).values.T).diagonal(),
                                 index=df.index
                                )
        anomalies = self.statistic[self.statistic>=self.ucl].index
        if show_figure:
            plt.figure()
            plt.plot(self.statistic,label='Hotelling statistic')
            plt.axhline(self.ucl,label='UCL',c='pink')
            for anom in anomalies:
                plt.axvline(anom,c='pink')
            plt.axvline(anom,c='pink',label=f'Anomalies, total {len(anomalies)} events')
            plt.xlabel('Datetime')
            plt.ylabel('Hotelling statistic')
            plt.title('Аномалии в нормальном режиме')
            plt.xticks(rotation=30)
            plt.legend()
            plt.show()            
        return anomalies
    
    
    def feature_importances(self,df):
        if not('ucl' in dir(self)):
            raise NameError("Fitting must be perfomed")
        feat_impor = []
        for col in df:
            _df = df.copy()
            _df[:] = 0
            _df[col] = (df - self.mean)[col]
            feat_impor.append(pd.Series(((_df.values @ self.inv_cov) @ _df.values.T).diagonal(),
                                        index=df.index) )
        feat_impor = pd.concat(feat_impor,1)#.rename(columns=df.columns)
        # нормировочка 
        _sum = feat_impor.sum(1).values
        for col in feat_impor:
            feat_impor[col] = (feat_impor[col].values / _sum) *100
        feat_impor.columns = df.columns
        return feat_impor
    
    def fit_predict(self,df,show_figure=False):
        self.fit(df)
        return self.predict(df,show_figure=show_figure)





from setuptools import setup, find_packages
 
setup(name='tsad',
      version='0.17.4',
      url='https://github.com/waico/tsad',
      license='BSD-3-Clause License',
      #packages=find_packages(),
      packages=['tsad'],
      author='Vyacheslav Kozitsin, Iurii Katser',
      author_email='rfptk2525@yandex.ru',
      description='Time Series Deep Learning Anomaly Detection',
      #packages=find_packages(exclude=['tests']),
      long_description=open('./tsad/README.md').read(),
      #long_description_content_type='text/x-rst',
      install_requires=['numpy>=1.19.5','pandas>=1.0.1','matplotlib>=3.1.3','scikit-learn>=0.24.1','torch==1.5.0'], # install_requires=[ 'A>=1,<2', 'B>=2']
      zip_safe=False)